/****************************************************************************
** Demonstration of drawing a Catmull-Rom spline, a curved line that       **
** passes through a number of control points.                              **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include "dos.h"
#include "stdlib.h"

//A pointer to the VGA memory
unsigned char far *Screen = (unsigned char far *)0xA0000000;
unsigned char far *Buffer;

//Number of points and the resolution while drawing
#define NumPts 5
#define Resolution 20

//A point consists of x and y coordinates
struct PointType {
  double x, y;
};

//Catmull-rom coefficients
struct PointType A, B, C, D;

//Control points}
struct PointType P [NumPts + 2];

//This sets the display to VGA 320x200 in 256 colours
void VGAScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x13;
  int86(0x10, &r, &r);
}

//This resets the display to text mode
void TextScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x3;
  int86(0x10, &r, &r);
}

//Sets a on the screen to a specific colour
void PutPixel (unsigned int x, unsigned int y, unsigned char c)
{
  if ((x < 320) && (y < 200))
    Buffer [(y << 8) + (y << 6) + x] = c;
}

//Draws a line between two points
void DrawLine (int x1, int y1, int x2, int y2, unsigned char c)
{
  int   i, deltax, deltay, numpixels,
        d, dinc1, dinc2,
        x, xinc1, xinc2,
        y, yinc1, yinc2;

  //Calculate deltax and deltay for initialisation
  deltax = abs (x2 - x1);
  deltay = abs (y2 - y1);

  //Initialize all vars based on which is the independent variable
  if (deltax >= deltay) {
    //x is independent variable
    numpixels = deltax + 1;
    d = (deltay << 1) - deltax;
    dinc1 = deltay << 1;
    dinc2 = (deltay - deltax) << 1;
    xinc1 = xinc2 = yinc2 = 1;
    yinc1 = 0;
  } else {
    //y is independent variable
    numpixels = deltay + 1;
    d = (deltax << 1) - deltay;
    dinc1 = deltax << 1;
    dinc2 = (deltax - deltay) << 1;
    xinc1 = 0;
    xinc2 = yinc1 = yinc2 = 1;
  }

  //Make sure x and y move in the right directions
  if (x1 > x2) {
    xinc1 = -xinc1;
    xinc2 = -xinc2;
  }
  if (y1 > y2) {
    yinc1 = - yinc1;
    yinc2 = - yinc2;
  }

  //Start drawing at <x1, y1>
  x = x1;
  y = y1;

  //Draw the pixels
  for (i = 0; i < numpixels; i++) {
    PutPixel (x, y, c);
    if (d < 0) {
      d += dinc1;
      x += xinc1;
      y += yinc1;
    } else {
      d += dinc2;
      x += xinc2;
      y += yinc2;
    }
  }
}

//Calculates one point on the spline
void SplinePoint (double t, struct PointType *Point)
{
  double t2, t3;

  //Square and cube of t
  t2 = t * t;
  t3 = t2 * t;

  //Calculate coordinates
  Point->x = ((A.x * t3) + (B.x * t2) + (C.x * t) + D.x) / 2;
  Point->y = ((A.y * t3) + (B.y * t2) + (C.y * t) + D.y) / 2;
}

/*
Computes coefficients for point n.
This is a matrix transform:
  -1  3 -3  1
   2 -5  4 -1
  -1  0  1  0
   0  2  0  0
*/
void ComputeCoeffs (int n)
{
  //x-coefficients
  A.x =    -P [n - 1].x + 3 * P [n].x - 3 * P [n + 1].x + P [n + 2].x;
  B.x = 2 * P [n - 1].x - 5 * P [n].x + 4 * P [n + 1].x - P [n + 2].x;
  C.x =    -P [n - 1].x                   + P [n + 1].x;
  D.x =                   2 * P [n].x;

  //y-coefficients
  A.y =    -P [n - 1].y + 3 * P [n].y - 3 * P [n + 1].y + P [n + 2].y;
  B.y = 2 * P [n - 1].y - 5 * P [n].y + 4 * P [n + 1].y - P [n + 2].y;
  C.y =    -P [n - 1].y                   + P [n + 1].y;
  D.y =                   2 * P [n].y;
}

//Draws the spline itself
void DrawSpline (int Points, unsigned char Colour)
{
  int             Point, Segment;
  struct PointType Current, Next;

  //Duplicate start and end points
  P [0]          = P [1];
  P [Points + 1] = P [Points];

  //Loop along all the points, drawing a line to the next point
  for (Point = 1; Point < Points; Point++) {
    //Calculate coefficients for this point
    ComputeCoeffs (Point);
    //Calculate the start point for the first segment
    SplinePoint (0, &Current);
    //Split into smaller segments
    for (Segment = 1; Segment <= Resolution; Segment++) {
      //Calculate end point
      SplinePoint ((double)Segment / Resolution, &Next);
      //Draw segment
      DrawLine (Current.x, Current.y, Next.x, Next.y, Colour);
      //Next part
      Current = Next;
    }
  }
}

void main ()
{
  unsigned char Point;

  randomize ();

  //Switch to VGA screen
  VGAScreen ();

  //I'm drawing to the screen instead of a buffer
  Buffer = Screen;

  do {
    //Clear buffer
    memset (Buffer, 0, 64000);

    //Get random points
    for (Point = 1; Point <= NumPts; Point++) {
      P [Point].x = random (320);
      P [Point].y = random (200);
    }

    //Draw the spline along those points
    DrawSpline (NumPts, 8);

    //Draw the points themselves - you'll probably want to leave this out
    for (Point = 1; Point <= NumPts; Point++)
      PutPixel (P [Point].x, P [Point].y, 7);

  } while (getch () != 27);

  TextScreen ();
}